import pkg_resources

from therapist.plugins import Plugin
from therapist.plugins.exc import InvalidPlugin, PluginNotInstalled


def list_plugins():
    return [entry_point.name for entry_point in pkg_resources.iter_entry_points(group="therapist.plugin")]


def load_plugin(name):
    for entry_point in pkg_resources.iter_entry_points(group="therapist.plugin", name=name):
        plugin = entry_point.load()

        if issubclass(plugin, Plugin):
            return plugin
        else:
            raise InvalidPlugin("Plugin must inherit from `Plugin`.")
    raise PluginNotInstalled("Plugin `{}` has not been installed.".format(name))



def parse_version(string):
    bits = str(string).split(".")
    while len(bits) < 3:
        bits.append(0)
    return [int(b) for b in bits[0:3]]


def version_compare(a, b):
    a = parse_version(a)
    b = parse_version(b)

    if a[0] > b[0]:
        return 1
    elif a[0] < b[0]:
        return -1
    elif a[1] > b[1]:
        return 1
    elif a[1] < b[1]:
        return -1
    elif a[2] > b[2]:
        return 1
    elif a[2] < b[2]:
        return -1
    else:
        return 0



import time

from six import iterkeys
from xml.etree import ElementTree

from therapist.collection import Collection
from therapist.process import Process


class Result(object):
    SUCCESS = 0
    FAILURE = 1
    ERROR = 2

    def __init__(self, process, status=None):
        self._process = None
        self.process = process
        self.status = status
        self.output = None
        self.error = None
        self.start_time = time.time()
        self.end_time = self.start_time
        self.modified_files = []

    def __str__(self):
        if self.is_success:
            return "SUCCESS"
        elif self.is_failure:
            return "FAILURE"
        elif self.is_error:
            return "ERROR"
        else:
            return "SKIP"

    def __repr__(self):  # pragma: no cover
        return "<Result {}>".format(self.process)

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, value):
        if not (isinstance(value, Process)):
            raise TypeError("Expected a `Process` object.")

        self._process = value

    @property
    def is_success(self):
        return self.status == self.SUCCESS

    @property
    def is_failure(self):
        return self.status == self.FAILURE

    @property
    def is_error(self):
        return self.status == self.ERROR

    @property
    def is_skip(self):
        return not (self.is_success or self.is_failure or self.is_error)

    @property
    def execution_time(self):
        return self.end_time - self.start_time

    def mark_complete(self, status=None, output=None, error=None):
        if status is not None:
            self.status = status
        self.output = output
        self.error = error
        self.end_time = time.time()

    @property
    def has_modified_files(self):
        return len(self.modified_files) > 0

    def add_modified_file(self, path):
        self.modified_files.append(path)


class ResultCollection(Collection):
    class Meta:
        object_class = Result

    @property
    def has_success(self):
        for result in self.objects:
            if result.is_success:
                return True
        return False

    @property
    def has_failure(self):
        for result in self.objects:
            if result.is_failure:
                return True
        return False

    @property
    def has_skip(self):
        for result in self.objects:
            if result.is_skip:
                return True
        return False

    @property
    def has_error(self):
        for result in self.objects:
            if result.is_error:
                return True
        return False

    @property
    def execution_time(self):
        return sum(result.execution_time for result in self.objects)

    @property
    def has_modified_files(self):
        for result in self.objects:
            if result.has_modified_files:
                return True
        return False

    @property
    def modified_files(self):
        modified_files = {}

        for result in self.objects:
            if result.has_modified_files:
                for path in result.modified_files:
                    if path not in modified_files:
                        modified_files[path] = []
                    modified_files[path].append(str(result.process))

        return [(path, modified_files[path]) for path in sorted(iterkeys(modified_files))]

    def count(self, **kwargs):
        if "status" in kwargs:
            return sum(1 for result in self.objects if result.status == kwargs.get("status"))
        return len(self.objects)

    def dump(self):
        """Returns the results in string format."""
        text = ""

        for result in self.objects:
            if result.is_failure or result.is_error:
                text += "\n#{red}#{bright}"
                text += "{}\n".format("".ljust(79, "="))

                status = "FAILED" if result.is_failure else "ERROR"
                text += "{}: {}\n".format(status, result.process)
                text += "{}\n#{{reset_all}}".format("".ljust(79, "="))

                if result.output:
                    text += result.output

                if result.error:
                    if result.output:
                        text += "\n{}\n".format("".ljust(79, "-"))
                        text += "Additional error output:\n"
                        text += "{}\n".format("".ljust(79, "-"))
                    text += result.error

                if not text.endswith("\n"):
                    text += "\n"

        if self.has_modified_files:
            text += "\n#{{yellow}}#{{bright}}{}\n".format("".ljust(79, "-"))
            text += "Modified files:\n"
            text += "{}\n".format("".ljust(79, "-"))
            for path, modified_by in self.modified_files:
                text += "#{{reset_all}}{} #{{cyan}}<- {}\n".format(path, ", ".join(modified_by))

        return text

    def dump_junit(self):
        """Returns a string containing XML mapped to the JUnit schema."""
        testsuites = ElementTree.Element(
            "testsuites",
            name="therapist",
            time=str(round(self.execution_time, 2)),
            tests=str(self.count()),
            failures=str(self.count(status=Result.FAILURE)),
            errors=str(self.count(status=Result.ERROR)),
        )

        for result in self.objects:
            failures = "1" if result.is_failure else "0"
            errors = "1" if result.is_error else "0"

            testsuite = ElementTree.SubElement(
                testsuites,
                "testsuite",
                id=result.process.name,
                name=str(result.process),
                time=str(round(result.execution_time, 2)),
                tests="1",
                failures=failures,
                errors=errors,
            )

            testcase = ElementTree.SubElement(testsuite, "testcase", time=str(round(result.execution_time, 2)))
            testcase.attrib["name"] = result.process.name

            if result.is_failure or result.is_error:
                if result.is_failure:
                    element = ElementTree.SubElement(testcase, "failure", type="failure")
                else:
                    element = ElementTree.SubElement(testcase, "error", type="error")

                if result.error:
                    element.text = result.error
                else:
                    element.text = result.output if result.output else ""

        xmlstr = ElementTree.tostring(testsuites, encoding="utf-8").decode("utf-8")
        return '<?xml version="1.0" encoding="UTF-8"?>\n{}'.format(xmlstr)



import subprocess

from therapist.collection import Collection
from therapist.exc import Error
from therapist.process import Process
from therapist.runner.result import Result


class Action(Process):
    def execute(self, **kwargs):
        files = self.filter_files(kwargs.get("files"))
        should_fix = kwargs.get("fix")

        result = Result(self)

        if should_fix and "fix" in self.config:
            command = self.config.get("fix")
        else:
            command = self.config.get("run")

        if command and files:
            command_with_args = command.format(files=" ".join(files))

            try:
                pipes = subprocess.Popen(
                    command_with_args, shell=True, cwd=kwargs.get("cwd"), stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
            except OSError as err:
                result.mark_complete(status=Result.ERROR, error="OSError {}".format(str(err)))
            else:
                std_out, std_err = pipes.communicate()

                status = Result.SUCCESS if pipes.returncode == 0 else Result.FAILURE
                result.mark_complete(status=status, output=std_out.decode("utf-8"), error=std_err.decode("utf-8"))

        return result


class ActionCollection(Collection):
    class Meta:
        object_class = Action

    class DoesNotExist(Error):
        pass

    def get(self, name):
        for action in self.objects:
            if action.name == name:
                return action
        raise self.DoesNotExist("`{}` is not a valid action.".format(name))



import os

from therapist.utils.git import Git, Status


class Runner(object):
    def __init__(self, cwd, files=None, **kwargs):
        # Options from kwargs:
        fix = kwargs.get("fix", False)

        # Git related options
        enable_git = kwargs.get("enable_git", False)
        include_unstaged = kwargs.get("include_unstaged", False)
        include_untracked = kwargs.get("include_untracked", False)
        include_unstaged_changes = kwargs.get("include_unstaged_changes", False)
        stage_modified_files = kwargs.get("stage_modified_files", False)

        self.cwd = os.path.abspath(cwd)
        self.unstaged_changes = False

        self.git = Git(repo_path=self.cwd) if enable_git else None
        self.include_unstaged_changes = include_unstaged_changes or include_unstaged
        self.fix = fix
        self.stage_modified_files = stage_modified_files
        self.file_mtimes = {}

        if files is None:
            files = []

            if self.git:
                untracked_files = "all" if include_untracked else "no"
                out, err, code = self.git.status(porcelain=True, untracked_files=untracked_files)

                for line in out.splitlines():
                    file_status = Status(line)

                    # Check if staged files were modified since being staged
                    if file_status.is_staged and file_status.is_modified:
                        self.unstaged_changes = True

                    # Skip unstaged files if the `unstaged` flag is False
                    if not file_status.is_staged and not include_unstaged and not include_untracked:
                        continue

                    # Skip deleted files
                    if file_status.is_deleted:
                        continue

                    files.append(file_status.path)

        for path in files:
            if os.path.exists(path):
                self.file_mtimes[path] = os.path.getmtime(path)

        self.files = files

    def run_process(self, process):
        """Runs a single action."""
        message = u"#{bright}"
        message += u"{} ".format(str(process)[:68]).ljust(69, ".")

        stashed = False
        if self.git and self.unstaged_changes and not self.include_unstaged_changes:
            out, err, code = self.git.stash(keep_index=True, quiet=True)
            stashed = code == 0

        try:
            result = process(files=self.files, cwd=self.cwd, fix=self.fix)

            # Check for modified files
            if self.git:
                out, err, code = self.git.status(porcelain=True, untracked_files="no")
                for line in out.splitlines():
                    file_status = Status(line)

                    # Make sure the file is one of the files that was processed
                    if file_status.path in self.files and file_status.is_modified:
                        mtime = os.path.getmtime(file_status.path) if os.path.exists(file_status.path) else 0
                        if mtime > self.file_mtimes.get(file_status.path, 0):
                            self.file_mtimes[file_status.path] = mtime
                            result.add_modified_file(file_status.path)
                            if self.stage_modified_files:
                                self.git.add(file_status.path)
            else:
                for path in self.files:
                    mtime = os.path.getmtime(path) if os.path.exists(path) else 0
                    if mtime > self.file_mtimes.get(path, 0):
                        result.add_modified_file(path)

        except:  # noqa: E722
            raise
        finally:
            if stashed and self.git:
                self.git.reset(hard=True, quiet=True)
                self.git.stash.pop(index=True, quiet=True)

        if result.is_success:
            message += u" #{green}[SUCCESS]"
        elif result.is_failure:
            message += u" #{red}[FAILURE]"
        elif result.is_skip:
            message += u" #{cyan}[SKIPPED]"
        elif result.is_error:
            message += u" #{red}[ERROR!!]"

        return result, message



import re


class Status(object):
    def __init__(self, status):
        self.state = status[0]
        self.is_modified = status[1] == "M"

        if self.is_renamed or self.is_copied:
            matches = re.search(r"(\S+?)\s+->\s+(\S+?)$", status[3:])
            self.original_path = matches.groups()[0]
            self.path = matches.groups()[1]
        else:
            self.path = status[3:]

    def __str__(self):
        status = "{:1}{:1}".format(self.state, "M" if self.is_modified else "")
        if self.is_renamed or self.is_copied:
            status = "{:3}{} -> {}".format(status, self.original_path, self.path)
        else:
            status = "{:3}{}".format(status, self.path)
        return status

    def __repr__(self):  # pragma: no cover
        return "<Status {}>".format(self.path)

    @property
    def is_staged(self):
        """If the state is empty then the file is unstaged."""
        return self.state not in (" ", "?")

    @property
    def is_untracked(self):
        """Returns true if the file is untracked."""
        return self.state == "?"

    @property
    def is_renamed(self):
        """Returns true if the file is renamed."""
        return self.state == "R"

    @property
    def is_copied(self):
        """Returns true if the file is copied."""
        return self.state == "C"

    @property
    def is_deleted(self):
        """Returns true if the file is deleted."""
        return self.state == "D"

    @property
    def is_added(self):
        """Returns true if the file is added."""
        return self.state == "A"



import os
import re
import shutil
import subprocess

import click
import colorama

from six import iteritems, print_

from therapist import __version__
from therapist.config import Config
from therapist.messages import (
    CONFIRM_PRESERVE_LEGACY_HOOK_MSG,
    CONFIRM_REPLACE_HOOK_MSG,
    CONFIRM_RESTORE_LEGACY_HOOK_MSG,
    CONFIRM_UNINSTALL_HOOK_MSG,
    COPYING_HOOK_MSG,
    COPYING_LEGACY_HOOK_MSG,
    CURRENT_HOOK_NOT_THERAPIST_MSG,
    DONE_COPYING_HOOK_MSG,
    DONE_COPYING_LEGACY_HOOK_MSG,
    DONE_INSTALLING_HOOK_MSG,
    DONE_REMOVING_LEGACY_HOOK_MSG,
    DONE_UNINSTALLING_HOOK_MSG,
    EXISTING_HOOK_MSG,
    HOOK_ALREADY_INSTALLED_MSG,
    INSTALLING_HOOK_MSG,
    INSTALL_ABORTED_MSG,
    LEGACY_HOOK_EXISTS_MSG,
    MISCONFIGURED_MSG,
    NOT_GIT_REPO_MSG,
    NO_HOOK_INSTALLED_MSG,
    NO_THERAPIST_CONFIG_FILE_MSG,
    REMOVING_LEGACY_HOOK_MSG,
    UNINSTALLING_HOOK_MSG,
    UNINSTALL_ABORTED_MSG,
    UNSTAGED_CHANGES_MSG,
    UPGRADE_HOOK_MSG,
)
from therapist.plugins.loader import list_plugins
from therapist.runner import Runner
from therapist.runner.result import ResultCollection
from therapist.utils.filesystem import current_git_dir, current_root, list_files
from therapist.utils.hook import calculate_hook_hash, read_hook_hash, read_hook_version
from therapist.utils.git import Git


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
HOOK_VERSION = 2

git = Git()


def output(message, **kwargs):
    def repl(match):  # pragma: no cover
        attr = match.group(0)[2:-1].upper()
        if hasattr(colorama.Fore, attr):
            return getattr(colorama.Fore, attr)
        elif hasattr(colorama.Style, attr):
            return getattr(colorama.Style, attr)
        else:
            return match.group(0)

    message, count = re.subn("#{(.+?)}", repl, message)
    message = colorama.Style.RESET_ALL + message + colorama.Style.RESET_ALL
    print_(message, **kwargs)


def report_misconfigured_and_exit(err):
    output(MISCONFIGURED_MSG.format(err.message))

    if err.code == Config.Misconfigured.PLUGIN_NOT_INSTALLED:
        output("Installed plugins:")
        for p in list_plugins():
            output(p)

    exit(1)


def get_config(disable_git=False):
    git_dir = current_git_dir()
    root_dir = current_root()
    extra_kw = {}

    git_root = os.path.dirname(git_dir) if git_dir else None

    if root_dir is None:
        output(NO_THERAPIST_CONFIG_FILE_MSG)
        exit(1)

    try:
        config = Config(root_dir)
    except Config.Misconfigured as err:
        report_misconfigured_and_exit(err)
    else:
        if not disable_git and git_root == root_dir:
            extra_kw["enable_git"] = True

    return config, extra_kw


@click.group(invoke_without_command=True)
@click.option("--version", "-V", is_flag=True, help="Show the version and exit.")
def cli(version):
    """A smart pre-commit hook for git."""
    if version:
        output("v{}".format(__version__))


@cli.command()
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force installation of the hook. This will replace any existing hook "
    "unless you also use the --preserve-legacy option.",
)
@click.option("--fix", is_flag=True, help="The hook will automatically fix problems where possible.")
@click.option("--no-stage-modified-files", is_flag=True, help="Disables the staging of files modified by the hook.")
@click.option("--no-color", is_flag=True, help="Disables colors and other rich output.")
@click.option("--preserve-legacy", is_flag=True, help="Preserves any existing pre-commit hook.")
def install(**kwargs):
    """Install the pre-commit hook."""
    force = kwargs.get("force")
    preserve_legacy = kwargs.get("preserve_legacy")

    colorama.init(strip=kwargs.get("no_color"))

    stdout = subprocess.check_output("which therapist", shell=True)
    therapist_bin = stdout.decode("utf-8").split()[0]

    git_dir = current_git_dir()

    if git_dir is None:
        output(NOT_GIT_REPO_MSG)
        exit(1)

    hook_options = {
        "fix": "--fix" if kwargs.get("fix") else "",
        "stage_modified_files": "" if kwargs.get("no_stage_modified_files") else "--stage-modified-files",
        "therapist_bin": therapist_bin,
    }

    srchook_path = os.path.join(BASE_DIR, "hooks", "pre-commit-template")
    with open(srchook_path, "r") as f:
        srchook = f.read()
    srchook_hash = calculate_hook_hash(srchook_path, hook_options)

    dsthook_path = os.path.join(git_dir, "hooks", "pre-commit")

    if os.path.isfile(dsthook_path):
        dsthook_hash = read_hook_hash(dsthook_path)
        if dsthook_hash:
            if dsthook_hash == srchook_hash:
                output(HOOK_ALREADY_INSTALLED_MSG)
                exit(0)
        else:
            if not force and not preserve_legacy:
                print(EXISTING_HOOK_MSG)
                preserve_legacy = click.confirm(CONFIRM_PRESERVE_LEGACY_HOOK_MSG, default=True)

            if preserve_legacy:
                output(COPYING_HOOK_MSG, end="")
                shutil.copy2(dsthook_path, "{}.legacy".format(dsthook_path))
                output(DONE_COPYING_HOOK_MSG)
            elif not force:
                if not click.confirm(CONFIRM_REPLACE_HOOK_MSG, default=False):
                    output(INSTALL_ABORTED_MSG)
                    exit(1)

    output(INSTALLING_HOOK_MSG, end="")
    with open(dsthook_path, "w+") as f:
        srchook = srchook.replace("%hash%", srchook_hash)
        for k, v in iteritems(hook_options):
            srchook = srchook.replace("%{}%".format(k), v)
        f.write(srchook)
    os.chmod(dsthook_path, 0o775)
    output(DONE_INSTALLING_HOOK_MSG)


@cli.command()
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="Force uninstallation of the Therapist pre-commit hook. This will "
    "also remove any legacy hook unless you also use the "
    "--restore-legacy option.",
)
@click.option("--no-color", is_flag=True, help="Disables colors and other rich output.")
@click.option("--restore-legacy", is_flag=True, help="Restores any legacy pre-commit hook.")
def uninstall(**kwargs):
    """Uninstall the current pre-commit hook."""
    force = kwargs.get("force")
    restore_legacy = kwargs.get("restore_legacy")

    colorama.init(strip=kwargs.get("no_color"))

    git_dir = current_git_dir()

    if git_dir is None:
        output(NOT_GIT_REPO_MSG)
        exit(1)

    hook_path = os.path.join(git_dir, "hooks", "pre-commit")

    if not os.path.isfile(hook_path):
        output(NO_HOOK_INSTALLED_MSG)
        exit(0)

    hook_hash = read_hook_hash(hook_path)

    if hook_hash:
        if not force:
            if not click.confirm(CONFIRM_UNINSTALL_HOOK_MSG, default=False):
                output(UNINSTALL_ABORTED_MSG)
                exit(1)
    else:
        output(CURRENT_HOOK_NOT_THERAPIST_MSG)
        exit(1)

    legacy_hook_path = os.path.join(git_dir, "hooks", "pre-commit.legacy")

    if os.path.isfile(legacy_hook_path):
        if not force and not restore_legacy:
            output(LEGACY_HOOK_EXISTS_MSG)
            restore_legacy = click.confirm(CONFIRM_RESTORE_LEGACY_HOOK_MSG, default=True)
        if restore_legacy:
            output(COPYING_LEGACY_HOOK_MSG, end="")
            shutil.copy2(legacy_hook_path, hook_path)
            os.remove(legacy_hook_path)
            output(DONE_COPYING_LEGACY_HOOK_MSG)
            exit(0)
        else:
            if force or click.confirm("Would you like to remove the legacy hook?", default=False):
                output(REMOVING_LEGACY_HOOK_MSG, end="")
                os.remove(legacy_hook_path)
                output(DONE_REMOVING_LEGACY_HOOK_MSG)

    output(UNINSTALLING_HOOK_MSG, end="")
    os.remove(hook_path)
    output(DONE_UNINSTALLING_HOOK_MSG)


@cli.command()
@click.argument("paths", nargs=-1)
@click.option("--action", "-a", default=None, help="A name of a specific action to be run.")
@click.option("--disable-git", is_flag=True, help="Disable git-aware features.")
@click.option("--enable-git", is_flag=True, help="Enable git-aware features.")
@click.option("--fix", is_flag=True, help="Automatically fixes problems where possible.")
@click.option("--include-unstaged", is_flag=True, help="Include unstaged files.")
@click.option("--include-unstaged-changes", is_flag=True, help="Include unstaged changes to staged files.")
@click.option("--include-untracked", is_flag=True, help="Include untracked files.")
@click.option("--junit-xml", default=None, help="Create a junit-xml style report file at the given path.")
@click.option("--no-color", is_flag=True, help="Disables colors and other rich output.")
@click.option("--plugin", "-p", default=None, help="A name of a specific plugin to be run.")
@click.option("--stage-modified-files", is_flag=True, help="Files that are modified by any actions should be staged.")
@click.option("--use-tracked-files", is_flag=True, help="Runs actions against all tracked files.")
@click.option("--quiet", "-q", is_flag=True, help="Suppress all output, unless an error occurs.")
def run(**kwargs):
    """Run the Therapist suite."""
    paths = kwargs.pop("paths", ())
    action = kwargs.pop("action")
    plugin = kwargs.pop("plugin")
    junit_xml = kwargs.pop("junit_xml")
    use_tracked_files = kwargs.pop("use_tracked_files")
    quiet = kwargs.pop("quiet")
    disable_git = kwargs.pop("disable_git")

    colorama.init(strip=kwargs.pop("no_color"))

    git_dir = current_git_dir()
    root_dir = current_root()
    config, extra_kw = get_config(disable_git=disable_git)
    kwargs.update(extra_kw)

    # Validate any installed hook is the minimum required version
    if git_dir:
        hook_path = os.path.join(git_dir, "hooks", "pre-commit")
        if os.path.isfile(hook_path):
            hook_version = read_hook_version(hook_path)
            if hook_version and hook_version < HOOK_VERSION:
                output(UPGRADE_HOOK_MSG)
                exit(1)

    files = []
    if paths:
        # We want to look at files in their current state if paths are passed through
        kwargs["include_unstaged_changes"] = True

        # If paths were provided get all the files for each path
        for path in paths:
            for f in list_files(path):
                f = os.path.relpath(f, root_dir)
                if not f.startswith(".."):  # Don't include files outside the repo root.
                    files.append(f)
    elif use_tracked_files:
        # If the use tracked files flag was passed, get a list of all the tracked files
        out, err, code = git.ls_files()
        files = out.splitlines()

        if kwargs.get("include_untracked"):
            out, err, code = git.ls_files(o=True, exclude_standard=True)
            files += out.splitlines()

    if files or paths:
        kwargs["files"] = files

    runner = Runner(config.cwd, **kwargs)
    results = ResultCollection()

    if runner.unstaged_changes and not quiet:
        output(UNSTAGED_CHANGES_MSG, end="\n\n")

    processes = list(config.actions) + list(config.plugins)
    processes.sort(key=lambda x: x.name)  # Sort the list of processes for consistent results

    if plugin:
        try:
            processes = [config.plugins.get(plugin)]
        except config.plugins.DoesNotExist as e:
            output("{}\nAvailable plugins:".format(e.message))

            for p in config.plugins:
                output(p.name)
            exit(1)

    if action:
        try:
            processes = [config.actions.get(action)]
        except config.actions.DoesNotExist as e:
            output("{}\nAvailable actions:".format(e.message))

            for a in config.actions:
                output(a.name)
            exit(1)

    for process in processes:
        result, message = runner.run_process(process)
        results.append(result)

        if not quiet:
            output(message)

    if junit_xml:
        with open(junit_xml, "w+") as f:
            f.write("{}".format(results.dump_junit()))

    if not quiet:
        output(results.dump())
        output("#{{bright}}{}\nCompleted in: {}s".format("".ljust(79, "-"), round(results.execution_time, 2)))

    if results.has_error:
        exit(1)
    elif results.has_failure:
        exit(2)


@cli.command()
@click.argument("shortcut", nargs=1)
@click.argument("paths", nargs=-1)
@click.pass_context
def use(ctx, shortcut, paths):
    """Use a shortcut."""
    config, _ = get_config()

    try:
        use_shortcut = config.shortcuts.get(shortcut)

        while use_shortcut.extends is not None:
            base = config.shortcuts.get(use_shortcut.extends)
            use_shortcut = base.extend(use_shortcut)
    except config.shortcuts.DoesNotExist as err:
        output("{}\nAvailable shortcuts:".format(err.message))

        for s in config.shortcuts:
            output(s.name)
        exit(1)
    else:
        options = use_shortcut.options
        for flag in use_shortcut.flags:
            options[flag.replace("-", "_")] = True

        options_string = ""
        for k, v in sorted(iteritems(options)):
            options_string += " --{}".format(k.replace("_", "-"))
            if v is not True:
                options_string += " {}".format(v)

        output("#{{dim}}$ therapist run{} {}\n".format(options_string, " ".join(paths)))

        options["paths"] = paths
        ctx.invoke(run, **options)



NOT_GIT_REPO_MSG = "#{red}Not a git repository (or any of the parent directories)"

HOOK_ALREADY_INSTALLED_MSG = "The pre-commit hook has already been installed."

EXISTING_HOOK_MSG = (
    "#{yellow}There is an existing pre-commit hook.\n"
    "#{reset_all}Therapist can preserve this legacy hook and run it before the Therapist pre-commit "
    "hook."
)

CONFIRM_PRESERVE_LEGACY_HOOK_MSG = "Would you like to preserve this legacy hook?"

COPYING_HOOK_MSG = "Copying `pre-commit` to `pre-commit.legacy`...\t"

DONE_COPYING_HOOK_MSG = "#{green}#{bright}DONE"

CONFIRM_REPLACE_HOOK_MSG = "Do you want to replace this hook?"

INSTALL_ABORTED_MSG = "Installation aborted."

INSTALLING_HOOK_MSG = "Installing pre-commit hook...\t"

DONE_INSTALLING_HOOK_MSG = "#{green}#{bright}DONE"

NO_HOOK_INSTALLED_MSG = "There is no pre-commit hook currently installed."

UNINSTALL_ABORTED_MSG = "Uninstallation aborted."

CONFIRM_UNINSTALL_HOOK_MSG = "Are you sure you want to uninstall the current pre-commit hook?"

CURRENT_HOOK_NOT_THERAPIST_MSG = (
    "#{yellow}The current pre-commit hook is not the Therapist pre-commit hook.\n" "#{reset_all}Uninstallation aborted."
)

LEGACY_HOOK_EXISTS_MSG = "#{yellow}There is a legacy pre-commit hook present."

CONFIRM_RESTORE_LEGACY_HOOK_MSG = "Would you like to restore the legacy hook?"

COPYING_LEGACY_HOOK_MSG = "Copying `pre-commit.legacy` to `pre-commit`...\t"

DONE_COPYING_LEGACY_HOOK_MSG = "#{green}#{bright}DONE"

REMOVING_LEGACY_HOOK_MSG = "Removing `pre-commit.legacy`...\t"

DONE_REMOVING_LEGACY_HOOK_MSG = "#{green}#{bright}DONE"

UNINSTALLING_HOOK_MSG = "Uninstalling pre-commit hook...\t"

DONE_UNINSTALLING_HOOK_MSG = "#{green}#{bright}DONE"

MISCONFIGURED_MSG = "#{{red}}Misconfigured: {}"

UNSTAGED_CHANGES_MSG = "#{yellow}You have unstaged changes."

NO_THERAPIST_CONFIG_FILE_MSG = "#{red}No Therapist configuration file was found."

UPGRADE_HOOK_MSG = (
    "#{red}The installed pre-commit hook is incompatible with the current version of Therapist.\n"
    "#{reset_all}Install the latest pre-commit hook by running `therapist install`."
)



"""
A lint suite runner with Git-aware capabilities and a smart pre-commit hook.

Full documentation for this tool is available at:
http://therapist.readthedocs.io/
"""
import os
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install


DEPENDENCIES = ["click >= 6.7", "colorama >= 0.3.7", "pathspec >= 0.5.0", "PyYAML >= 3.12", "six >= 1.10.0"]

ROOT = os.path.abspath(os.path.dirname(__file__))

version = __import__("therapist").__version__


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != "v{}".format(version):
            info = "Git tag: {0} does not match the version of this app: {1}".format(tag, version)
            sys.exit(info)


setup(
    name="therapist",
    version=version,
    url="https://github.com/rehandalal/therapist",
    license="Mozilla Public License Version 2.0",
    author="Rehan Dalal",
    author_email="rehan@meet-rehan.com",
    description="A lint suite runner with Git-aware capabilities and a smart pre-commit hook.",
    long_description=__doc__,
    packages=find_packages(exclude=["tests", "docs"]),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    install_requires=DEPENDENCIES,
    py_modules=["therapist"],
    entry_points={"console_scripts": ["therapist = therapist.cli:cli"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    cmdclass={"verify": VerifyVersionCommand},
)



from therapist.collection import Collection
from therapist.exc import Error


class Shortcut(object):
    def __init__(self, name, **kwargs):
        self.name = name
        self.extends = kwargs.pop("extends", None)

        self._options = None
        self.options = kwargs.pop("options", None)

        self._flags = None
        self.flags = kwargs.pop("flags", None)

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, value):
        if value is None:
            self._options = {}
        elif isinstance(value, dict):
            self._options = {}
            for key in value:
                self._options[key.replace("-", "_")] = value[key]
        else:
            self.options = {value.replace("-", "_"): True}

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, value):
        if value is None:
            self._flags = set()
        elif isinstance(value, set):
            self._flags = value
        elif isinstance(value, list):
            self._flags = set(value)
        else:
            self._flags = set([value])

    def extend(self, apply):
        options = {}
        options.update(self.options)
        options.update(apply.options)

        flags = self.flags
        flags = flags.union(apply.flags)

        return Shortcut(apply.name, extends=self.extends, options=options, flags=flags)


class ShortcutCollection(Collection):
    class Meta:
        object_class = Shortcut

    class DoesNotExist(Error):
        pass

    def get(self, name):
        for shortcut in self.objects:
            if shortcut.name == name:
                return shortcut
        raise self.DoesNotExist("`{}` is not a valid shortcut.".format(name))



import os
import yaml

from therapist import __version__
from therapist.exc import Error
from therapist.plugins.exc import InvalidPlugin, PluginNotInstalled
from therapist.plugins.loader import load_plugin
from therapist.plugins.plugin import PluginCollection
from therapist.runner.action import Action, ActionCollection
from therapist.runner.shortcut import Shortcut, ShortcutCollection
from therapist.utils import parse_version, version_compare


class Config(object):
    class Misconfigured(Error):
        NO_CONFIG_FILE = 1
        NO_ACTIONS_OR_PLUGINS = 2
        EMPTY_CONFIG = 3
        ACTIONS_WRONGLY_CONFIGURED = 4
        PLUGINS_WRONGLY_CONFIGURED = 5
        PLUGIN_NOT_INSTALLED = 6
        PLUGIN_INVALID = 7
        SHORTCUTS_WRONGLY_CONFIGURED = 8
        VERSION_MISMATCH = 9
        REQUIRES_WRONGLY_CONFIGURED = 10

        def __init__(self, *args, **kwargs):
            self.code = kwargs.pop("code", None)
            super(self.__class__, self).__init__(*args, **kwargs)

    def __init__(self, cwd):
        self.cwd = os.path.abspath(cwd)
        self.actions = ActionCollection()
        self.plugins = PluginCollection()
        self.shortcuts = ShortcutCollection()
        self.requires = None

        # Try and load the config file
        try:
            with open(os.path.join(self.cwd, ".therapist.yml"), "r") as f:
                config = yaml.safe_load(f)
        except IOError:
            raise self.Misconfigured("Missing configuration file.", code=self.Misconfigured.NO_CONFIG_FILE)
        else:
            if config is None:
                raise self.Misconfigured("Empty configuration file.", code=self.Misconfigured.EMPTY_CONFIG)

            if "actions" in config:
                try:
                    actions = config["actions"]
                except TypeError:
                    raise self.Misconfigured(
                        "`actions` was not configured correctly.", code=self.Misconfigured.ACTIONS_WRONGLY_CONFIGURED
                    )
                else:
                    if not isinstance(actions, dict):
                        raise self.Misconfigured(
                            "`actions` was not configured correctly.",
                            code=self.Misconfigured.ACTIONS_WRONGLY_CONFIGURED,
                        )

                    for action_name in actions:
                        settings = actions[action_name]
                        if settings is None:
                            settings = {}
                        self.actions.append(Action(action_name, **settings))

            if "plugins" in config:
                try:
                    plugins = config["plugins"]
                except TypeError:
                    raise self.Misconfigured(
                        "`plugins` was not configured correctly.", code=self.Misconfigured.PLUGINS_WRONGLY_CONFIGURED
                    )
                else:
                    if not isinstance(plugins, dict):
                        raise self.Misconfigured(
                            "`plugins` was not configured correctly.",
                            code=self.Misconfigured.PLUGINS_WRONGLY_CONFIGURED,
                        )

                    for plugin_name in plugins:
                        try:
                            plugin = load_plugin(plugin_name)
                        except PluginNotInstalled:
                            raise self.Misconfigured(
                                "Plugin `{}` is not installed.".format(plugin_name),
                                code=self.Misconfigured.PLUGIN_NOT_INSTALLED,
                            )
                        except InvalidPlugin:
                            raise self.Misconfigured(
                                "Plugin `{}` is not a valid plugin.".format(plugin_name),
                                code=self.Misconfigured.PLUGIN_INVALID,
                            )
                        settings = plugins[plugin_name]
                        if settings is None:
                            settings = {}
                        self.plugins.append(plugin(plugin_name, **settings))

            if "shortcuts" in config:
                try:
                    shortcuts = config["shortcuts"]
                except TypeError:
                    raise self.Misconfigured(
                        "`shortcuts` was not configured correctly.",
                        code=self.Misconfigured.SHORTCUTS_WRONGLY_CONFIGURED,
                    )
                else:
                    if not isinstance(shortcuts, dict):
                        raise self.Misconfigured(
                            "`shortcuts` was not configured correctly.",
                            code=self.Misconfigured.SHORTCUTS_WRONGLY_CONFIGURED,
                        )

                    for shortcut_name in shortcuts:
                        settings = shortcuts[shortcut_name]
                        if settings is None:
                            settings = {}
                        self.shortcuts.append(Shortcut(shortcut_name, **settings))

            if "requires" in config:
                self.requires = config["requires"]

                if isinstance(self.requires, str) or isinstance(self.requires, float) or isinstance(self.requires, int):
                    self.requires = {"min": self.requires}
                elif not isinstance(self.requires, dict):
                    raise self.Misconfigured(
                        "`requires` must be of type string, float, int or dict.",
                        code=self.Misconfigured.REQUIRES_WRONGLY_CONFIGURED,
                    )

                min_version = self.requires.get("min")
                max_version = self.requires.get("max")

                mismatch_min = False
                mismatch_max = False

                try:
                    if min_version and max_version:
                        if version_compare(min_version, max_version) > 0:
                            raise self.Misconfigured(
                                "`min` value should be less than `max`.",
                                code=self.Misconfigured.REQUIRES_WRONGLY_CONFIGURED,
                            )
                    if min_version:
                        if version_compare(min_version, __version__) > 0:
                            mismatch_min = ".".join([str(v) for v in parse_version(min_version)])
                    if max_version:
                        if version_compare(max_version, __version__) < 0:
                            mismatch_max = ".".join([str(v) for v in parse_version(max_version)])
                except ValueError:
                    raise self.Misconfigured(
                        "Invalid version specified in `requires`.", code=self.Misconfigured.REQUIRES_WRONGLY_CONFIGURED
                    )

                if mismatch_min:
                    raise self.Misconfigured(
                        "You require Therapist >= {}.".format(mismatch_min), code=self.Misconfigured.VERSION_MISMATCH
                    )
                elif mismatch_max:
                    raise self.Misconfigured(
                        "You require Therapist <= {}.".format(mismatch_max), code=self.Misconfigured.VERSION_MISMATCH
                    )

            if not (self.actions or self.plugins):
                raise self.Misconfigured(
                    "`actions` or `plugins` must be specified in the configuration file.",
                    code=self.Misconfigured.NO_ACTIONS_OR_PLUGINS,
                )



import hashlib

from six import iterkeys


def read_hook_hash(path):
    """Verify that the file at path is the therapist hook and return the hash"""
    with open(path, "r") as f:
        f.readline()  # Discard the shebang line
        version_line = f.readline()
        if version_line.startswith("# THERAPIST"):
            return version_line.split()[2]


def read_hook_version(path):
    """Read the hook version from the file."""
    with open(path, "r") as f:
        f.readline()  # Discard the shebang line
        version_line = f.readline()
        if version_line.startswith("# THERAPIST"):
            try:
                return int(version_line.split()[3][1:])
            except IndexError:
                return 1


def calculate_hook_hash(path, options):
    """Hash a hook file"""
    with open(path, "r") as f:
        data = f.read()
        for key in sorted(iterkeys(options)):
            data += "\n#{}={}".format(key, options.get(key))
        return hashlib.md5(data.encode()).hexdigest()
